﻿using System;
using UnityEngine;
using UnityEngine.UI;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using MVR.FileManagementSecure;

/***********************************************************************************************************************************
 * prestigitis HeelAdjust 1.0.2
 * 
 * prestigitis_HeelAdjust.cs 
 * 
 * requires VaM 1.19.2.2
 * add this script to a Person atom to adjust foot pose and add heel and toe colliders for high-heeled or thick-soled shoes. 
 * autoload is triggered by mouse left button or controller button release while the 'clothing' or 'clothing preset' tab is active.
 * for oculus touch, autoload is triggered only when using touch controller button A.
 * OnDestroy() is not working, so disable the plugin before removing it.
 * 
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License
 * https://creativecommons.org/licenses/by-nc-sa/4.0/
 *
 ************************************************************************************************************************************
 * 1.0.0 (20200109) prestigitis
 *     - original release
 * 1.0.1 
 *     JayJayWon
 *     - fixed for VaM 1.19 changes in VaM clothing hierarchy
 *     - fixed for VaM 1.19 changes in clothing item naming convention
 *     - saved and unsaved active clothing checks combined into one method
 *     prestigitis
 *     - suppress repeated errors in FixedUpdate when Person UI has not been loaded
 *     - automagically create HeelAdjust folder if it does not exist
 * 1.0.2
 *     ICannotDie
 *     - added support for male models. See SetModelGender().
 ***********************************************************************************************************************************/

namespace prestigitis
{
    public class HeelAdjust : MVRScript
    {
        /***********************************************************************************************************************************/
        /* configuration settings 
        /***********************************************************************************************************************************/
        private bool AutoLoadEnabled = true;                             // determines whether auto load is initially enabled
        private float AutoLoadFootPoseDelay = 0.3f;                      // how long after clicking clothing tab UI before foot pose is autoloaded
        private string HeelAdjustSubfolderName = "HeelAdjust";           // name of subfolder in which heeladjust files are saved in Saves folder
        private string HeelAdjustFileExtension = "heeladjust";           // file extension for saved files
        private Color ShoeColliderColor = new Color(0.0f, 0f, 1f, 0.7f); // color of the toe collider when visible
        private float FCV3ResetDelay = 0.5f;                             // how long after adjustment until foot fcv3 rotation states reset
        private float ShoeColliderBaseScale = 0.01f;                     // initial vertical scale of shoe colliders. each slider unit of height = this many meters.
        private float ToeColliderMinHeight = 0.01f;                      // minimum height of toe collider
        private float HeelColliderMinHeight = 0.01f;                     // minimum height of heel collider
        private float HeelColliderInitialBackwardOffset = -0.035f;       // how far backward heel collider is initially offset from foot transform
        /***********************************************************************************************************************************/

        // IMPORTANT - DO NOT make custom enums. The dynamic C# complier crashes Unity when it encounters these for
        // some reason

        // IMPORTANT - DO NOT OVERRIDE Awake() as it is used internally by MVRScript - instead use Init() function which
        // is called right after creation

        private bool FCV3SaveRotationStateEnable = true;
        protected JSONStorableFloat FootAdjust_X_JSON;
        protected JSONStorableFloat FootAdjust_Y_JSON;
        protected JSONStorableFloat FootAdjust_Z_JSON;
        protected JSONStorableFloat ToeAdjust_X_JSON;
        protected JSONStorableFloat ToeAdjust_Y_JSON;
        protected JSONStorableFloat ToeAdjust_Z_JSON;
        protected JSONStorableFloat HeelColliderHeight_JSON;
        protected JSONStorableFloat HeelColliderBackOffset_JSON;
        protected JSONStorableFloat ToeCollidersHeight_JSON;
        private JSONStorableString settingsOutputField;
        private JSONStorableBool CollidersVisible_JSON;
        private JSONStorableBool AutoLoadEnabled_JSON;
        private string settingsOutputFormat;
        private UIDynamicTextField _infoTextField;
        protected JSONStorableStringChooser clothingItem_JSON;
        private Vector3 ToeColliderBaseRotationR = new Vector3(-16.7f, 0f, -4.3f);
        private Vector3 ToeColliderBaseRotationL = new Vector3(-16.7f, 0f, 4.3f);
        private Vector3 HeelColliderBaseRotationR = new Vector3(-20f, 0f, 0f); //rFoot world rotation: Vector3(13.1f, 9.2f, 1.3f);
        private Vector3 HeelColliderBaseRotationL = new Vector3(-20f, 0f, 0f); //lFoot world rotation: Vector3(13.1f, 351.4f, 359.4f);
        private bool SlidersEnabled = false;
        private FreeControllerV3.RotationState FCV3RotationStateRFoot = FreeControllerV3.RotationState.On;
        private FreeControllerV3.RotationState FCV3RotationStateLFoot = FreeControllerV3.RotationState.On;
        private FreeControllerV3.RotationState FCV3RotationStateRToe = FreeControllerV3.RotationState.On;
        private FreeControllerV3.RotationState FCV3RotationStateLToe = FreeControllerV3.RotationState.On;
        private string _modelGender = "Female";

        protected void AdjustFootPoseCallback(JSONStorableFloat jf)
        {
            if (SlidersEnabled)
            {
                if (FCV3SaveRotationStateEnable) SaveFootFCV3RotationStates();
                Vector3 rf = new Vector3(FootAdjust_X_JSON.val, FootAdjust_Y_JSON.val, FootAdjust_Z_JSON.val);
                Vector3 rt = new Vector3(ToeAdjust_X_JSON.val - FootAdjust_X_JSON.val, ToeAdjust_Y_JSON.val, ToeAdjust_Z_JSON.val);
                Vector3 lf = new Vector3(FootAdjust_X_JSON.val, -FootAdjust_Y_JSON.val, -FootAdjust_Z_JSON.val);
                Vector3 lt = new Vector3(ToeAdjust_X_JSON.val - FootAdjust_X_JSON.val, -ToeAdjust_Y_JSON.val, -ToeAdjust_Z_JSON.val);
                SetToeCollidersXRotation(-ToeAdjust_X_JSON.val); //compensate for toe rotation to keep toe colliders flat
                SetHeelCollidersXRotation(-FootAdjust_X_JSON.val);
                AdjustFootPose(rf, rt, lf, lt);
                UpdateTextField();
                StartCoroutine(FCV3AutoReset());
            }
        }
        protected void ColliderHeight_Callback(JSONStorableFloat jf)
        {
            //adjust collider height
            if (SlidersEnabled)
            {
                AdjustShoeCollider();
                UpdateTextField();
            }
        }
        protected void CollidersVisibleCallback(bool b)
        {
            Transform rt = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescale/rShoeToeCollider");
            Transform lt = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lToe/lShoeToeColliderRescale/lShoeToeCollider");
            Transform rh = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescale/rShoeHeelCollider");
            Transform lh = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lShoeHeelColliderRescale/lShoeHeelCollider");

            rt.GetComponent<Renderer>().enabled = b;
            lt.GetComponent<Renderer>().enabled = b;
            rh.GetComponent<Renderer>().enabled = b;
            lh.GetComponent<Renderer>().enabled = b;
        }
        protected void AutoLoadEnabledCallback(bool b)
        {
            AutoLoadEnabled = b;
        }
        protected bool Load(string path)
        {
            string loadedSettings = SuperController.singleton.ReadFileIntoString(path);
            string[] loadedSettingsLines = loadedSettings.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.None);
            string[] footVals = loadedSettingsLines[0].Split(new string[] { "," }, StringSplitOptions.None);
            string[] toeVals = loadedSettingsLines[1].Split(new string[] { "," }, StringSplitOptions.None);
            string[] heelColliderVals = loadedSettingsLines[2].Split(new string[] { "," }, StringSplitOptions.None);
            string[] toeColliderVals = loadedSettingsLines[3].Split(new string[] { "," }, StringSplitOptions.None);
            FootAdjust_X_JSON.val = float.Parse(footVals[0]);
            FootAdjust_Y_JSON.val = float.Parse(footVals[1]);
            FootAdjust_Z_JSON.val = float.Parse(footVals[2]);
            ToeAdjust_X_JSON.val = float.Parse(toeVals[0]);
            ToeAdjust_Y_JSON.val = float.Parse(toeVals[1]);
            ToeAdjust_Z_JSON.val = float.Parse(toeVals[2]);
            HeelColliderHeight_JSON.val = float.Parse(heelColliderVals[0]);
            HeelColliderBackOffset_JSON.val = float.Parse(heelColliderVals[1]);
            ToeCollidersHeight_JSON.val = float.Parse(toeColliderVals[0]);
            return true;
        }
        protected bool SaveNew(string path)
        {
            //save path must be in 'Saves' folder...
            if (clothingItem_JSON.val == "None") return false;
            else
            {
                SuperController.singleton.SaveStringIntoFile(path, settingsOutputField.val);
                SyncClothingChoices();
                if (ActiveClothes(true).Count > 0) clothingItem_JSON.val = ActiveClothes(true).First<string>();
                else clothingItem_JSON.val = "None";
                return true;
            }
        }
        protected bool SaveOver(string path)
        {
            //save path must be in 'Saves' folder
            SuperController.singleton.SaveStringIntoFile(path, settingsOutputField.val);
            return true;
        }
        private void SelectClothingCallback(string s)
        {
            //nothing yet
        }
        protected void SyncClothingChoices()
        {
            clothingItem_JSON.choices = ActiveClothes(true);
            if (!clothingItem_JSON.choices.Contains(clothingItem_JSON.val)) clothingItem_JSON.val = "None";
        }
        public override void Init()
        {
            try
            {
                pluginLabelJSON.val = "prestigitis HeelAdjust 1.0.2";

                // folder check
                if (!FileManagerSecure.DirectoryExists(SuperController.singleton.savesDir + HeelAdjustSubfolderName))
                {
                    SuperController.LogMessage("Creating directory \"[VaM]\\" + SuperController.singleton.savesDir + HeelAdjustSubfolderName + "\\\"");
                    FileManagerSecure.CreateDirectory(SuperController.singleton.savesDir + HeelAdjustSubfolderName);
                }

                FootAdjust_X_JSON = new JSONStorableFloat("Feet X", 0f, AdjustFootPoseCallback, -90f, 90f, true);
                RegisterFloat(FootAdjust_X_JSON);
                CreateSlider(FootAdjust_X_JSON);

                FootAdjust_Y_JSON = new JSONStorableFloat("Feet Y", 0f, AdjustFootPoseCallback, -60f, 60f, true);
                RegisterFloat(FootAdjust_Y_JSON);
                CreateSlider(FootAdjust_Y_JSON);

                FootAdjust_Z_JSON = new JSONStorableFloat("Feet Z", 0f, AdjustFootPoseCallback, -60f, 60f, true);
                RegisterFloat(FootAdjust_Z_JSON);
                CreateSlider(FootAdjust_Z_JSON);

                ToeAdjust_X_JSON = new JSONStorableFloat("Toes X", 0f, AdjustFootPoseCallback, -90f, 90f, true);
                RegisterFloat(ToeAdjust_X_JSON);
                CreateSlider(ToeAdjust_X_JSON);

                ToeAdjust_Y_JSON = new JSONStorableFloat("Toes Y", 0f, AdjustFootPoseCallback, -20f, 20f, true);
                RegisterFloat(ToeAdjust_Y_JSON);
                CreateSlider(ToeAdjust_Y_JSON);

                ToeAdjust_Z_JSON = new JSONStorableFloat("Toes Z", 0f, AdjustFootPoseCallback, -20f, 20f, true);
                RegisterFloat(ToeAdjust_Z_JSON);
                CreateSlider(ToeAdjust_Z_JSON);

                CollidersVisible_JSON = new JSONStorableBool("Colliders Visible", false, new JSONStorableBool.SetBoolCallback(CollidersVisibleCallback));
                RegisterBool(CollidersVisible_JSON);
                CreateToggle(CollidersVisible_JSON);

                HeelColliderHeight_JSON = new JSONStorableFloat("Heel Collider Height", 0.010f, ColliderHeight_Callback, 0.0f, 20.0f, false);
                RegisterFloat(HeelColliderHeight_JSON);
                CreateSlider(HeelColliderHeight_JSON);

                HeelColliderBackOffset_JSON = new JSONStorableFloat("Heel Backward Offset", -0.030f, ColliderHeight_Callback, -0.10f, 0f, false);
                RegisterFloat(HeelColliderBackOffset_JSON);
                CreateSlider(HeelColliderBackOffset_JSON);

                ToeCollidersHeight_JSON = new JSONStorableFloat("Toe Collider Height", 0.010f, ColliderHeight_Callback, 0.0f, 10.0f, false);
                RegisterFloat(ToeCollidersHeight_JSON);
                CreateSlider(ToeCollidersHeight_JSON);

                CreateButton("Reset foot pose", true).button.onClick.AddListener(() => {
                    FootPoseReset();
                });

                var infoBuilder = new StringBuilder();

                infoBuilder.AppendLine("[prestigitis] HeelAdjust 1.0.1");
                infoBuilder.AppendLine();
                infoBuilder.AppendLine("Foot pose adjustments and colliders for shoes");
                infoBuilder.AppendLine();
                infoBuilder.AppendLine("Editable contents of *." + HeelAdjustFileExtension + " file are shown below:");

                _infoTextField = CreateTextField(new JSONStorableString("info", infoBuilder.ToString()), true);
                _infoTextField.height = 300;
                _infoTextField.UItext.fontSize = 32;

                StringTextbox(ref settingsOutputField, "." + HeelAdjustFileExtension + " file data", "", _ => { settingsOutputFormat = settingsOutputField.val; }, true);

                CreateButton("Load *." + HeelAdjustFileExtension + " file...", true).button.onClick.AddListener(() => {
                    SuperController.singleton.GetMediaPathDialog((string path) => {
                        Load(path);
                    }, HeelAdjustFileExtension, SuperController.singleton.savesDir + "/" + HeelAdjustSubfolderName, true);
                });

                AutoLoadEnabled_JSON = new JSONStorableBool("Auto load enabled", true, new JSONStorableBool.SetBoolCallback(AutoLoadEnabledCallback));
                RegisterBool(AutoLoadEnabled_JSON);
                CreateToggle(AutoLoadEnabled_JSON, true);

                CreateButton("Overwrite existing *." + HeelAdjustFileExtension + " file...", true).button.onClick.AddListener(() => {
                    SuperController.singleton.GetMediaPathDialog((string path) => {
                        SaveOver(path);
                    }, HeelAdjustFileExtension, SuperController.singleton.savesDir + "/" + HeelAdjustSubfolderName, false);
                });

                clothingItem_JSON = new JSONStorableStringChooser(
                    "clothing item", ActiveClothes(true), "None",
                    "clothing item for new save", new JSONStorableStringChooser.SetStringCallback(SelectClothingCallback));
                clothingItem_JSON.popupOpenCallback = new JSONStorableStringChooser.PopupOpenCallback(SyncClothingChoices);
                RegisterStringChooser(clothingItem_JSON);
                CreateScrollablePopup(clothingItem_JSON, true);

                CreateButton("Save new *." + HeelAdjustFileExtension + " file", true).button.onClick.AddListener(() => {
                    SaveNew(SuperController.singleton.savesDir + "/" + HeelAdjustSubfolderName + "/" + clothingItem_JSON.val + "." + HeelAdjustFileExtension);
                });
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        // Start is called once before Update or FixedUpdate is called and after Init()
        void Start()
        {
            try
            {
                SetModelGender();
                AddToeColliders();
                AddHeelColliders();
                CollidersVisibleCallback(false);
                SaveFootFCV3RotationStates();
                UpdateSlidersFromFootPose();
                SlidersEnabled = true;
                LoadFootPoseFromActiveClothing();
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        // Update is called with each rendered frame by Unity
        void Update()
        {
            try
            {

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        // FixedUpdate is called with each physics simulation frame by Unity
        void FixedUpdate()
        {
            try
            {
                if (containingAtom.gameObject.transform.Find("UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Clothing") && containingAtom.gameObject.transform.Find("UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Clothing Presets"))
                {
                    var cUItab = containingAtom.gameObject.transform.Find("UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Clothing").GetComponent<UITab>();
                    var cpUItab = containingAtom.gameObject.transform.Find("UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Clothing Presets").GetComponent<UITab>();
                    if (cUItab && cpUItab)
                    {
                        if (cUItab.isActiveAndEnabled || cpUItab.isActiveAndEnabled)
                        {
                            if (Input.GetMouseButtonUp(0) || Input.GetButtonUp(SuperController.singleton.buttonSelect))
                            {
                                StartCoroutine(AutoLoadFootPose());
                                SyncClothingChoices();
                            }
                        }
                    }

                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        void OnEnable()
        {
            if (FCV3SaveRotationStateEnable) SaveFootFCV3RotationStates();
            StartCoroutine(FCV3AutoReset());

            //apply heel rotations + collider enable and size
            Vector3 rf = new Vector3(FootAdjust_X_JSON.val, FootAdjust_Y_JSON.val, FootAdjust_Z_JSON.val);
            Vector3 rt = new Vector3(ToeAdjust_X_JSON.val - FootAdjust_X_JSON.val, ToeAdjust_Y_JSON.val, ToeAdjust_Z_JSON.val);
            Vector3 lf = new Vector3(FootAdjust_X_JSON.val, -FootAdjust_Y_JSON.val, -FootAdjust_Z_JSON.val);
            Vector3 lt = new Vector3(ToeAdjust_X_JSON.val - FootAdjust_X_JSON.val, -ToeAdjust_Y_JSON.val, -ToeAdjust_Z_JSON.val);
            SetToeCollidersXRotation(-ToeAdjust_X_JSON.val); //compensate for toe rotation to keep toe colliders flat
            SetHeelCollidersXRotation(-FootAdjust_X_JSON.val);
            AdjustFootPose(rf, rt, lf, lt);
            UpdateTextField();
            AdjustShoeCollider();
            LoadFootPoseFromActiveClothing();
            CollidersVisibleCallback(CollidersVisible_JSON.val);
        }
        void OnDisable()
        {
            //reset foot and toe rotations to their defaults
            Vector3 rfootBaseRotation = new Vector3(0f, 0f, 0f);
            Vector3 rtoeBaseRotation = new Vector3(0f, 0f, 0f);
            Vector3 lfootBaseRotation = new Vector3(0f, 0f, 0f);
            Vector3 ltoeBaseRotation = new Vector3(0f, 0f, 0f);

            var rfoot = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot");
            var rtoe = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rToe");
            var lfoot = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot");
            var ltoe = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lToe");

            rfoot.GetComponent<DAZBone>().baseJointRotation = rfootBaseRotation;
            rtoe.GetComponent<DAZBone>().baseJointRotation = rtoeBaseRotation;
            lfoot.GetComponent<DAZBone>().baseJointRotation = lfootBaseRotation;
            ltoe.GetComponent<DAZBone>().baseJointRotation = ltoeBaseRotation;

            SetToeCollidersHeight(ToeColliderMinHeight);
            SetHeelCollidersHeight(HeelColliderMinHeight);
            CollidersVisibleCallback(false);
        }
        void OnDestroy()
        {
            // OnDestroy is where you should put any cleanup
            // if you registered objects to supercontroller or atom, you should unregister them here
        }

        private void SetModelGender()
        {
            _modelGender = containingAtom.GetComponentInChildren<DAZCharacterSelector>()?.gender == DAZCharacterSelector.Gender.Male ? "Male" : "Female";
        }

        void FootFCV3RotationEnable(bool e)
        {
            //turn off rotation for free controllers while adjusting
            var rf = containingAtom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/rToeControl").GetComponent<FreeControllerV3>();
            var rt = containingAtom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/rFootControl").GetComponent<FreeControllerV3>();
            var lf = containingAtom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/lToeControl").GetComponent<FreeControllerV3>();
            var lt = containingAtom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/lFootControl").GetComponent<FreeControllerV3>();
            if (e)
            {
                rf.currentRotationState = FCV3RotationStateRFoot;
                rt.currentRotationState = FCV3RotationStateRToe;
                lf.currentRotationState = FCV3RotationStateLFoot;
                lt.currentRotationState = FCV3RotationStateLToe;
            }
            else
            {
                rf.currentRotationState = FreeControllerV3.RotationState.Off;
                rt.currentRotationState = FreeControllerV3.RotationState.Off;
                lf.currentRotationState = FreeControllerV3.RotationState.Off;
                lt.currentRotationState = FreeControllerV3.RotationState.Off;
            }
        }
        void SaveFootFCV3RotationStates()
        {
            var rf = containingAtom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/rToeControl").GetComponent<FreeControllerV3>();
            var rt = containingAtom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/rFootControl").GetComponent<FreeControllerV3>();
            var lf = containingAtom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/lToeControl").GetComponent<FreeControllerV3>();
            var lt = containingAtom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/lFootControl").GetComponent<FreeControllerV3>();

            FCV3RotationStateRFoot = rf.currentRotationState;
            FCV3RotationStateRToe = rt.currentRotationState;
            FCV3RotationStateLFoot = lf.currentRotationState;
            FCV3RotationStateLToe = lt.currentRotationState;
        }
        IEnumerator FCV3AutoReset()
        {
            FootFCV3RotationEnable(false);
            FCV3SaveRotationStateEnable = false;
            yield return new WaitForSeconds(FCV3ResetDelay);
            FootFCV3RotationEnable(true);
            FCV3SaveRotationStateEnable = true;
        }
        void FootPoseReset()
        {
            //reset sliders to zero/minimum values
            Vector3 footBaseRotation = new Vector3(0f, 0f, 0f);
            Vector3 toeBaseRotation = new Vector3(0f, 0f, 0f);
            FootAdjust_X_JSON.val = footBaseRotation.x;
            FootAdjust_Y_JSON.val = footBaseRotation.y;
            FootAdjust_Z_JSON.val = footBaseRotation.z;
            ToeAdjust_X_JSON.val = toeBaseRotation.x;
            ToeAdjust_Y_JSON.val = toeBaseRotation.y;
            ToeAdjust_Z_JSON.val = toeBaseRotation.z;
            HeelColliderHeight_JSON.val = HeelColliderMinHeight;
            HeelColliderBackOffset_JSON.val = HeelColliderInitialBackwardOffset;
            ToeCollidersHeight_JSON.val = ToeColliderMinHeight;
        }
        void AdjustFootPose(Vector3 rFootAngle, Vector3 rToeAngle, Vector3 lFootAngle, Vector3 lToeAngle)
        {
            //adjust foot pose to the given angles
            var rfoot = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot");
            var rtoe = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rToe");
            var lfoot = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot");
            var ltoe = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lToe");

            rfoot.GetComponent<DAZBone>().baseJointRotation = rFootAngle;
            rtoe.GetComponent<DAZBone>().baseJointRotation = rToeAngle;
            lfoot.GetComponent<DAZBone>().baseJointRotation = lFootAngle;
            ltoe.GetComponent<DAZBone>().baseJointRotation = lToeAngle;
        }
        void UpdateSlidersFromFootPose()
        {
            //update sliders from the right foot only
            //foot pose
            var rfoot = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot");
            var rtoe = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rToe");

            FootAdjust_X_JSON.val = rfoot.GetComponent<DAZBone>().baseJointRotation.x;
            FootAdjust_Y_JSON.val = rfoot.GetComponent<DAZBone>().baseJointRotation.y;
            FootAdjust_Z_JSON.val = rfoot.GetComponent<DAZBone>().baseJointRotation.z;

            ToeAdjust_X_JSON.val = rtoe.GetComponent<DAZBone>().baseJointRotation.x + rfoot.GetComponent<DAZBone>().baseJointRotation.x;
            ToeAdjust_Y_JSON.val = rtoe.GetComponent<DAZBone>().baseJointRotation.y;
            ToeAdjust_Z_JSON.val = rtoe.GetComponent<DAZBone>().baseJointRotation.z;

            //colliders
            var rbc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rShoeCollider").GetComponent<BoxCollider>();
            HeelColliderHeight_JSON.val = rbc.size.z;

            Transform rh = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescale/rShoeHeelCollider");
            HeelColliderBackOffset_JSON.val = rh.transform.localPosition.z;

            Transform rtc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescale");
            ToeCollidersHeight_JSON.val = rtc.transform.localScale.y;// = new Vector3(1f, h, 1f);
        }
        void AdjustShoeCollider()
        {
            //adjust/enable shoe collider from slider values
            if (HeelColliderHeight_JSON.val > HeelColliderMinHeight)
            {
                SetHeelCollidersHeight(HeelColliderHeight_JSON.val);
            }
            else
            {
                SetHeelCollidersHeight(HeelColliderMinHeight);
                HeelColliderHeight_JSON.val = HeelColliderMinHeight;
            }
            if (ToeCollidersHeight_JSON.val > ToeColliderMinHeight)
            {
                SetToeCollidersHeight(ToeCollidersHeight_JSON.val);
            }
            else
            {
                SetToeCollidersHeight(ToeColliderMinHeight);
                ToeCollidersHeight_JSON.val = ToeColliderMinHeight;
            }
            SetHeelColliderBackOffset(HeelColliderBackOffset_JSON.val);
        }
        void AddToeColliders()
        {
            Transform rt = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rToe");
            Transform lt = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lToe");

            GameObject rShoeToeColliderRescale = new GameObject();
            rShoeToeColliderRescale.transform.position = rt.position;
            rShoeToeColliderRescale.transform.rotation = rt.rotation;
            rShoeToeColliderRescale.transform.parent = rt;
            rShoeToeColliderRescale.transform.localEulerAngles = ToeColliderBaseRotationR;
            rShoeToeColliderRescale.name = "rShoeToeColliderRescale";

            GameObject rShoeToeCollider = GameObject.CreatePrimitive(PrimitiveType.Cube);
            rShoeToeCollider.transform.position = rShoeToeColliderRescale.transform.position;
            rShoeToeCollider.transform.parent = rShoeToeColliderRescale.transform;
            rShoeToeCollider.transform.localScale = new Vector3(0.05f, ShoeColliderBaseScale, 0.05f);
            rShoeToeCollider.name = "rShoeToeCollider";
            rShoeToeCollider.GetComponent<Collider>().enabled = true;
            rShoeToeCollider.transform.localPosition = new Vector3(0f, -ShoeColliderBaseScale / 2, 0f);
            rShoeToeCollider.transform.localEulerAngles = new Vector3(0f, 0.0f, 0.0f);
            rShoeToeCollider.GetComponent<Renderer>().material.shader = Shader.Find("Transparent/Diffuse");
            rShoeToeCollider.GetComponent<Renderer>().material.color = ShoeColliderColor;

            GameObject lShoeToeColliderRescale = new GameObject();
            lShoeToeColliderRescale.transform.position = lt.position;
            lShoeToeColliderRescale.transform.rotation = lt.rotation;
            lShoeToeColliderRescale.transform.parent = lt;
            lShoeToeColliderRescale.transform.localEulerAngles = ToeColliderBaseRotationL;
            lShoeToeColliderRescale.name = "lShoeToeColliderRescale";

            GameObject lShoeToeCollider = GameObject.CreatePrimitive(PrimitiveType.Cube);
            lShoeToeCollider.transform.position = lShoeToeColliderRescale.transform.position;
            lShoeToeCollider.transform.parent = lShoeToeColliderRescale.transform;
            lShoeToeCollider.transform.localScale = new Vector3(0.05f, ShoeColliderBaseScale, 0.05f);
            lShoeToeCollider.name = "lShoeToeCollider";
            lShoeToeCollider.GetComponent<Collider>().enabled = true;
            lShoeToeCollider.transform.localPosition = new Vector3(0f, -ShoeColliderBaseScale / 2, 0f);
            lShoeToeCollider.transform.localEulerAngles = new Vector3(0f, 0.0f, 0.0f);
            lShoeToeCollider.GetComponent<Renderer>().material.shader = Shader.Find("Transparent/Diffuse");
            lShoeToeCollider.GetComponent<Renderer>().material.color = ShoeColliderColor;

            //retract initially
            rShoeToeColliderRescale.transform.localScale = new Vector3(1f, ToeColliderMinHeight, 1f);
            lShoeToeColliderRescale.transform.localScale = new Vector3(1f, ToeColliderMinHeight, 1f);
        }
        void AddHeelColliders()
        {
            Transform rh = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot");
            Transform lh = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot");

            GameObject rShoeHeelColliderRescale = new GameObject();
            rShoeHeelColliderRescale.transform.position = rh.position;
            rShoeHeelColliderRescale.transform.rotation = rh.rotation;
            rShoeHeelColliderRescale.transform.parent = rh;
            rShoeHeelColliderRescale.transform.localEulerAngles = HeelColliderBaseRotationR;
            rShoeHeelColliderRescale.name = "rShoeHeelColliderRescale";

            GameObject rShoeHeelCollider = GameObject.CreatePrimitive(PrimitiveType.Cube);
            rShoeHeelCollider.transform.position = new Vector3(rh.position.x, rh.position.y, rh.position.z - HeelColliderInitialBackwardOffset * 50);
            rShoeHeelCollider.transform.parent = rShoeHeelColliderRescale.transform;
            rShoeHeelCollider.transform.localScale = new Vector3(0.02f, ShoeColliderBaseScale, 0.02f);
            rShoeHeelCollider.name = "rShoeHeelCollider";
            rShoeHeelCollider.GetComponent<Collider>().enabled = true;
            rShoeHeelCollider.transform.localPosition = new Vector3(0f, -ShoeColliderBaseScale / 2, HeelColliderInitialBackwardOffset);
            rShoeHeelCollider.transform.localEulerAngles = new Vector3(0f, 0.0f, 0.0f);
            rShoeHeelCollider.GetComponent<Renderer>().material.shader = Shader.Find("Transparent/Diffuse");
            rShoeHeelCollider.GetComponent<Renderer>().material.color = ShoeColliderColor;

            GameObject lShoeHeelColliderRescale = new GameObject();
            lShoeHeelColliderRescale.transform.position = lh.position;
            lShoeHeelColliderRescale.transform.rotation = lh.rotation;
            lShoeHeelColliderRescale.transform.parent = lh;
            lShoeHeelColliderRescale.transform.localEulerAngles = HeelColliderBaseRotationL;
            lShoeHeelColliderRescale.name = "lShoeHeelColliderRescale";

            GameObject lShoeHeelCollider = GameObject.CreatePrimitive(PrimitiveType.Cube);
            lShoeHeelCollider.transform.parent = lShoeHeelColliderRescale.transform;
            lShoeHeelCollider.transform.localScale = new Vector3(0.02f, ShoeColliderBaseScale, 0.02f);
            lShoeHeelCollider.name = "lShoeHeelCollider";
            lShoeHeelCollider.GetComponent<Collider>().enabled = true;
            lShoeHeelCollider.transform.localPosition = new Vector3(0f, -ShoeColliderBaseScale / 2, HeelColliderInitialBackwardOffset);
            lShoeHeelCollider.transform.localEulerAngles = new Vector3(0f, 0.0f, 0.0f);
            lShoeHeelCollider.GetComponent<Renderer>().material.shader = Shader.Find("Transparent/Diffuse");
            lShoeHeelCollider.GetComponent<Renderer>().material.color = ShoeColliderColor;

            //retract initially
            rShoeHeelColliderRescale.transform.localScale = new Vector3(1f, HeelColliderMinHeight, 1f);
            lShoeHeelColliderRescale.transform.localScale = new Vector3(1f, HeelColliderMinHeight, 1f);
        }
        void SetToeCollidersHeight(float h)
        {
            Transform rc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescale");
            Transform lc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lToe/lShoeToeColliderRescale");
            rc.transform.localScale = new Vector3(1f, h, 1f);
            lc.transform.localScale = new Vector3(1f, h, 1f);
        }
        void SetHeelCollidersHeight(float h)
        {
            Transform rc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescale");
            Transform lc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lShoeHeelColliderRescale");
            rc.transform.localScale = new Vector3(1f, h, 1f);
            lc.transform.localScale = new Vector3(1f, h, 1f);
        }
        void SetToeCollidersXRotation(float r)
        {
            Transform rc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescale");
            Transform lc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lToe/lShoeToeColliderRescale");
            rc.transform.localEulerAngles = new Vector3(r + ToeColliderBaseRotationR.x, ToeColliderBaseRotationR.y, ToeColliderBaseRotationR.z);
            lc.transform.localEulerAngles = new Vector3(r + ToeColliderBaseRotationL.x, ToeColliderBaseRotationL.y, ToeColliderBaseRotationL.z);
        }
        void SetHeelCollidersXRotation(float r)
        {
            Transform rc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescale");
            Transform lc = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lShoeHeelColliderRescale");
            rc.transform.localEulerAngles = new Vector3(r + HeelColliderBaseRotationR.x, HeelColliderBaseRotationR.y, HeelColliderBaseRotationR.z);
            lc.transform.localEulerAngles = new Vector3(r + HeelColliderBaseRotationL.x, HeelColliderBaseRotationL.y, HeelColliderBaseRotationL.z);
        }
        void SetHeelColliderBackOffset(float f)
        {
            Transform rh = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescale/rShoeHeelCollider");
            Transform lh = containingAtom.gameObject.transform.Find($"rescale2/PhysicsModel/Genesis2{_modelGender}/hip/pelvis/lThigh/lShin/lFoot/lShoeHeelColliderRescale/lShoeHeelCollider");
            rh.transform.localPosition = new Vector3(0f, -ShoeColliderBaseScale / 2, f);
            lh.transform.localPosition = new Vector3(0f, -ShoeColliderBaseScale / 2, f);
        }
        private void StringTextbox(ref JSONStorableString output, string name, string start, JSONStorableString.SetStringCallback callback, bool rhs)
        {
            output = new JSONStorableString(name, start, callback);

            RegisterString(output);
            var textfield = CreateTextField(output, rhs);
            var input = textfield.gameObject.AddComponent<InputField>();
            input.textComponent = textfield.UItext;
            textfield.backgroundColor = Color.white;
            input.lineType = InputField.LineType.MultiLineNewline;
            input.textComponent.resizeTextForBestFit = true;
            output.inputField = input;
        }
        private void UpdateTextField()
        {
            string s = FootAdjust_X_JSON.val.ToString() + "," + FootAdjust_Y_JSON.val.ToString() + "," + FootAdjust_Z_JSON.val.ToString() + "\n";
            s = s + ToeAdjust_X_JSON.val.ToString() + "," + ToeAdjust_Y_JSON.val.ToString() + "," + ToeAdjust_Z_JSON.val.ToString() + "\n";
            s = s + HeelColliderHeight_JSON.val.ToString() + "," + HeelColliderBackOffset_JSON.val.ToString() + "\n";
            s = s + ToeCollidersHeight_JSON.val.ToString();
            settingsOutputField.val = s;
        }
        private string GetItemNameFromCustomClothingUID(string uid)
        {
            string itemName = uid;
            if (uid.Contains("/"))
            {
                itemName = uid.Substring(uid.LastIndexOf("/") + 1);
                if (uid.EndsWith(".vam"))
                {
                    itemName = itemName.Substring(0, itemName.Length - 4);
                }
            }
            return (itemName);
        }
        private List<string> ActiveClothes(bool unsavedOnly)
        {
            //create list of active clothing items
            List<string> activeUnsavedClothes = new List<string>();
            foreach (Transform child in containingAtom.gameObject.transform.Find($"rescale2/geometry/{_modelGender}Clothes"))
            {
                if (child.gameObject.name == $"{_modelGender}ClothingPrefab(Clone)")
                {
                    foreach (Transform subchild in containingAtom.gameObject.transform.Find($"rescale2/geometry/{_modelGender}Clothes/{_modelGender}ClothingPrefab(Clone)"))
                    {
                        if (subchild.GetComponent<DAZClothingItem>().isActiveAndEnabled)
                        {
                            string s = StripNonAlphaNum(subchild.GetComponent<DAZClothingItem>().creatorName) + "-" + StripNonAlphaNum(subchild.GetComponent<DAZClothingItem>().uid);

                            if (!ExistingFiles().Contains(s) || !unsavedOnly) activeUnsavedClothes.Add(s);
                        }
                    }
                }
                else
                {
                    if (child.GetComponent<DAZClothingItem>().isActiveAndEnabled)
                    {
                        string s = StripNonAlphaNum(child.GetComponent<DAZClothingItem>().creatorName) + "-" + StripNonAlphaNum(GetItemNameFromCustomClothingUID(child.GetComponent<DAZClothingItem>().uid));

                        if (!ExistingFiles().Contains(s) || !unsavedOnly) activeUnsavedClothes.Add(s);
                    }
                }
            }

            return activeUnsavedClothes;
        }
        private List<string> ExistingFiles()
        {
            //list of files in heeladjust dir
            List<string> files = new List<string>();
            foreach (string file in SuperController.singleton.GetFilesAtPath(SuperController.singleton.savesDir + HeelAdjustSubfolderName))
            {
                string[] pathsplit = file.Split(new string[] { "\\" }, StringSplitOptions.None);
                files.Add(pathsplit.Last<String>().Replace("." + HeelAdjustFileExtension, ""));
            }
            return files;
        }
        private string StripNonAlphaNum(string s)
        {
            return Regex.Replace(s, @"[^A-Za-z0-9]+", "");
        }
        IEnumerator AutoLoadFootPose()
        {
            yield return new WaitForSeconds(AutoLoadFootPoseDelay);
            LoadFootPoseFromActiveClothing();
        }
        private bool LoadFootPoseFromActiveClothing()
        {
            if (!AutoLoadEnabled) return false;
            //get list of active clothing items
            List<string> activeClothes = ActiveClothes(false);

            //get intersection of active clothing and files
            List<string> loadableProfiles = activeClothes.Intersect(ExistingFiles()).ToList();

            if (loadableProfiles.Count < 1) //intersection list is empty
            {
                FootPoseReset();
                return false;
            }
            else
            {
                //a heeladjust file was found for one of the active clothing items, load the first one
                Load(SuperController.singleton.savesDir + "/" + HeelAdjustSubfolderName + "/" + loadableProfiles.First().ToString() + "." + HeelAdjustFileExtension);
                return true;
            }
        }
    }
}

